package be.iminds.ilabt.jfed.preferences;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.GDPRInfo;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Testbed;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.JFedUtils;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Singleton;
import java.io.File;
import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.util.*;

/**
 * Manages a list of for which testbeds data processing consent settings have been handled by the user
 */
@Singleton
public class TestbedGdprConsentPreferences {
    private final static Logger LOG = LoggerFactory.getLogger(TestbedGdprConsentPreferences.class);

    private class ConsentKey {
        protected final String url; //GDPR consent URL is used as ID, not testbed. (multiple testbeds can use same URL)
        protected final String version;
        protected final GDPRInfo.AcceptSubject subjectType;
        protected final String subject; //user URN, project URN or slice URN as string

        public ConsentKey(String url, String version, GDPRInfo.AcceptSubject subjectType, String subject) {
            this.url = url;
            this.version = version;
            this.subjectType = subjectType;
            this.subject = subject;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            ConsentKey that = (ConsentKey) o;
            return Objects.equals(url, that.url) &&
                    Objects.equals(version, that.version) &&
                    subjectType == that.subjectType &&
                    Objects.equals(subject, that.subject);
        }

        @Override
        public int hashCode() {
            //expire is ignored for equality and hashcode!
            return Objects.hash(url, version,subjectType, subject);
        }
    }

    /**
     * helper for storing consent as json
     */
    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(ignoreUnknown = true)
    private class ConsentObj extends ConsentKey {
        private final Instant expire;

        public ConsentObj(@JsonProperty("url") String url,
                          @JsonProperty("version") String version,
                          @JsonProperty("subjectType") GDPRInfo.AcceptSubject subjectType,
                          @JsonProperty("subject") String subject,
                          @JsonProperty("expire") Instant expire) {
            super(url, version, subjectType, subject);
            this.expire = expire;
        }

        public ConsentObj(@Nonnull ConsentKey k, Instant expire) {
            super(k.url, k.version, k.subjectType, k.subject);
            this.expire = expire;
        }

        @JsonProperty public Instant getExpire() { return expire; }
        @JsonProperty public String getUrl() { return url; }
        @JsonProperty public String getVersion() { return version; }
        @JsonProperty public GDPRInfo.AcceptSubject getSubjectType() { return subjectType; }
        @JsonProperty public String getSubject() { return subject; }
    }

    private final Map<ConsentKey, Instant> storedConsentExpiration;

    public TestbedGdprConsentPreferences() {
        this.storedConsentExpiration = new HashMap<>();
        try {
            File consentPrefFile = JFedUtils.createPreferencesFile("gdpr-consent.json");
            ObjectMapper mapper = Jackson.newObjectMapper();
            List<ConsentObj> consentObjs = mapper.readValue(consentPrefFile, mapper.getTypeFactory().constructCollectionType(List.class, ConsentObj.class));
            if (consentObjs != null) {
                for (ConsentObj consentObj : consentObjs) {
                    storedConsentExpiration.put(consentObj, consentObj.getExpire());
                }
            }
        } catch (IOException e) {
            LOG.warn("Error restoring GDPR Consent preferences", e);
        }
    }

    private void save() {
        try {
            File consentPrefFile = JFedUtils.createPreferencesFile("gdpr-consent.json");
            ObjectMapper mapper = Jackson.newObjectMapper();
            List<ConsentObj> consentObjs = new ArrayList<>();
            for (Map.Entry<ConsentKey, Instant> e : storedConsentExpiration.entrySet()) {
                ConsentKey k = e.getKey();
                Instant expire = e.getValue();
                consentObjs.add(new ConsentObj(k, expire));
            }
            mapper.writerWithDefaultPrettyPrinter().writeValue(consentPrefFile, consentObjs);
        } catch (IOException e) {
            LOG.warn("Error saving GDPR Consent preferences", e);
        }
    }

    @Nullable
    private ConsentKey createConsentKey(@Nonnull Testbed testbed, @Nonnull String projectName,
                                        @Nonnull GeniUrn userUrn/*, @Nonnull GeniUrn sliceUrn*/) {
        GDPRInfo gdprInfo = testbed.getGdprInfo();
        if (gdprInfo == null)
            return null;
        if (gdprInfo.getGrpdUrl() == null)
            return null;
        if (gdprInfo.getAcceptSubject() == null)
            return null;

        String subject;
        switch (gdprInfo.getAcceptSubject()) {
            case USER: { subject = userUrn.toString(); break; }
            case PROJECT: { subject = projectName; break; }
//            case SLICE: { subject = sliceUrn.toString(); break; }
            default: throw new RuntimeException("Unsupported AcceptSubject: "+gdprInfo.getAcceptSubject());
        }
        return new ConsentKey(gdprInfo.getGrpdUrl(), gdprInfo.getVersion(),
                gdprInfo.getAcceptSubject(), subject);
    }

    public boolean isConsentManaged(@Nonnull Testbed testbed, @Nonnull String projectName,
                                    @Nonnull GeniUrn userUrn/*, @Nonnull GeniUrn sliceUrn*/) {
        ConsentKey consentKey = createConsentKey(testbed, projectName, userUrn/*, sliceUrn*/);
        if (consentKey == null)
            return true;
        Instant expire = storedConsentExpiration.get(consentKey);
        if (expire == null)
            return false;
        return expire.isAfter(Instant.now().plus(Duration.ofMinutes(10)));
    }

    /**
     * Mark consent as managed (using now() date)
     */
    public void markConsentManaged(@Nonnull Testbed testbed, @Nonnull String projectName,
                                   @Nonnull GeniUrn userUrn/*, @Nonnull GeniUrn sliceUrn*/) {
        if (testbed.getGdprInfo() == null)
            return;
        if (testbed.getGdprInfo().getAcceptPeriodInDays() == null)
            return;

        ConsentKey consentKey = createConsentKey(testbed, projectName, userUrn/*, sliceUrn*/);
        if (consentKey == null)
            return;

        storedConsentExpiration.put(consentKey, Instant.now().plus(Duration.ofDays(testbed.getGdprInfo().getAcceptPeriodInDays())));

        save();
    }
}
